@@ -7,20 +7,21 @@
 # Copyright (C) 2008 Gary van der Merwe
 # Copyright (C) 2008 Hendrik van Antwerpen
 # Copyright (C) 2008 ojnkpjg
-# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2008-2011, 2014, 2018-2021 Philipp Wolfer
 # Copyright (C) 2009 Nikolai Prokoschenko
 # Copyright (C) 2011-2012 Chad Wilson
 # Copyright (C) 2011-2013, 2019 Michael Wiencek
 # Copyright (C) 2012-2013, 2016-2017 Wieland Hoffmann
 # Copyright (C) 2013, 2018 Calvin Walton
 # Copyright (C) 2013-2015, 2017 Sophist-UK
-# Copyright (C) 2013-2015, 2017-2019 Laurent Monin
+# Copyright (C) 2013-2015, 2017-2021 Laurent Monin
 # Copyright (C) 2016 Suhas
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2017 Antonio Larrosa
 # Copyright (C) 2018 Vishal Choudhary
 # Copyright (C) 2019 Joel Lintunen
-# Copyright (C) 2020 Gabriel Ferreira
+# Copyright (C) 2020-2021 Gabriel Ferreira
+# Copyright (C) 2021 Petit Minion
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -42,6 +43,7 @@ from collections import (
     defaultdict,
     namedtuple,
 )
+from enum import IntEnum
 import traceback
 
 from PyQt5 import (
@@ -49,12 +51,10 @@ from PyQt5 import (
     QtNetwork,
 )
 
-from picard import (
-    config,
-    log,
-)
+from picard import log
 from picard.cluster import Cluster
 from picard.collection import add_release_to_user_collections
+from picard.config import get_config
 from picard.const import VARIOUS_ARTISTS_ID
 from picard.dataobj import DataObject
 from picard.file import File
@@ -111,9 +111,16 @@ class AlbumArtist(DataObject):
         super().__init__(album_artist_id)
 
 
+class AlbumStatus(IntEnum):
+    NONE = 0
+    LOADING = 1
+    ERROR = 2
+    LOADED = 3
+
+
 class Album(DataObject, Item):
 
-    release_group_loaded = QtCore.pyqtSignal()
+    metadata_images_changed = QtCore.pyqtSignal()
 
     def __init__(self, album_id, discid=None):
         DataObject.__init__(self, album_id)
@@ -123,7 +130,7 @@ class Album(DataObject, Item):
         self.loaded = False
         self.load_task = None
         self.release_group = None
-        self._files = 0
+        self._files_count = 0
         self._requests = 0
         self._tracks_loaded = False
         self._discids = set()
@@ -131,8 +138,8 @@ class Album(DataObject, Item):
             self._discids.add(discid)
         self._after_load_callbacks = []
         self.unmatched_files = Cluster(_("Unmatched Files"), special=True, related_album=self, hide_if_empty=True)
-        self.errors = []
-        self.status = None
+        self.unmatched_files.metadata_images_changed.connect(self.update_metadata_images)
+        self.status = AlbumStatus.NONE
         self._album_artists = []
         self.update_metadata_images_enabled = True
 
@@ -141,11 +148,9 @@ class Album(DataObject, Item):
 
     def iterfiles(self, save=False):
         for track in self.tracks:
-            for file in track.iterfiles():
-                yield file
+            yield from track.iterfiles()
         if not save:
-            for file in self.unmatched_files.iterfiles():
-                yield file
+            yield from self.unmatched_files.iterfiles()
 
     def enable_update_metadata_images(self, enabled):
         self.update_metadata_images_enabled = enabled
@@ -167,7 +172,7 @@ class Album(DataObject, Item):
             if track_discids:
                 track.metadata['musicbrainz_discid'] = track_discids
                 track.update()
-                for file in track.linked_files:
+                for file in track.files:
                     file.metadata['musicbrainz_discid'] = track_discids
                     file.update()
 
@@ -222,6 +227,8 @@ class Album(DataObject, Item):
         m.copy(rg.metadata)
         release_to_metadata(release_node, m, album=self)
 
+        config = get_config()
+
         # Custom VA name
         if m['musicbrainz_albumartistid'] == VARIOUS_ARTISTS_ID:
             m['albumartistsort'] = m['albumartist'] = config.setting['va_name']
@@ -254,6 +261,7 @@ class Album(DataObject, Item):
                 self.error_append(http.errorString())
                 # Fix for broken NAT releases
                 if error == QtNetwork.QNetworkReply.ContentNotFoundError:
+                    config = get_config()
                     nats = False
                     nat_name = config.setting["nat_name"]
                     files = list(self.unmatched_files.files)
@@ -276,21 +284,11 @@ class Album(DataObject, Item):
             self._requests -= 1
             if parsed or error:
                 self._finalize_loading(error)
-        # does http need to be set to None to free the memory used by the network response?
-        # http://qt-project.org/doc/qt-5/qnetworkaccessmanager.html says:
-        #     After the request has finished, it is the responsibility of the user
-        #     to delete the QNetworkReply object at an appropriate time.
-        #     Do not directly delete it inside the slot connected to finished().
-        #     You can use the deleteLater() function.
-
-    def error_append(self, msg):
-        log.error(msg)
-        self.errors.append(msg)
 
     def _finalize_loading(self, error):
         if error:
             self.metadata.clear()
-            self.status = _("[could not load album %s]") % self.id
+            self.status = AlbumStatus.ERROR
             del self._new_metadata
             del self._new_tracks
             self.update()
@@ -319,7 +317,6 @@ class Album(DataObject, Item):
                 mm = Metadata()
                 mm.copy(self._new_metadata)
                 medium_to_metadata(medium_node, mm)
-                discpregap = False
                 format = medium_node.get('format')
                 if format:
                     all_media.append(format)
@@ -327,29 +324,43 @@ class Album(DataObject, Item):
                 for dj in djmix_ars.get(mm["discnumber"], []):
                     mm.add("djmixer", dj)
 
+                if va:
+                    mm["compilation"] = "1"
+                else:
+                    del mm["compilation"]
+
                 if 'discs' in medium_node:
                     discids = [disc.get('id') for disc in medium_node['discs']]
                     mm['~musicbrainz_discids'] = discids
                     mm['musicbrainz_discid'] = list(self._discids.intersection(discids))
 
                 if "pregap" in medium_node:
-                    discpregap = True
                     absolutetracknumber += 1
-                    track = self._finalize_loading_track(medium_node['pregap'], mm, artists, va, absolutetracknumber, discpregap)
-                    track.metadata['~pregap'] = "1"
+                    mm['~discpregap'] = '1'
+                    extra_metadata = {
+                        '~pregap': '1',
+                        '~absolutetracknumber': absolutetracknumber,
+                    }
+                    self._finalize_loading_track(medium_node['pregap'], mm, artists, extra_metadata)
 
                 track_count = medium_node['track-count']
                 if track_count:
                     tracklist_node = medium_node['tracks']
                     for track_node in tracklist_node:
                         absolutetracknumber += 1
-                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
+                        extra_metadata = {
+                            '~absolutetracknumber': absolutetracknumber,
+                        }
+                        self._finalize_loading_track(track_node, mm, artists, extra_metadata)
 
                 if "data-tracks" in medium_node:
                     for track_node in medium_node['data-tracks']:
                         absolutetracknumber += 1
-                        track = self._finalize_loading_track(track_node, mm, artists, va, absolutetracknumber, discpregap)
-                        track.metadata['~datatrack'] = "1"
+                        extra_metadata = {
+                            '~datatrack': '1',
+                            '~absolutetracknumber': absolutetracknumber,
+                        }
+                        self._finalize_loading_track(track_node, mm, artists, extra_metadata)
 
             totalalbumtracks = absolutetracknumber
             self._new_metadata['~totalalbumtracks'] = totalalbumtracks
@@ -366,6 +377,10 @@ class Album(DataObject, Item):
 
         if not self._requests:
             self.enable_update_metadata_images(False)
+            for track in self._new_tracks:
+                track.orig_metadata.copy(track.metadata)
+                track.metadata_images_changed.connect(self.update_metadata_images)
+
             # Prepare parser for user's script
             for s_name, s_text in enabled_tagger_scripts_texts():
                 parser = ScriptParser()
@@ -384,19 +399,18 @@ class Album(DataObject, Item):
                     log.exception("Failed to run tagger script %s on album", s_name)
                 self._new_metadata.strip_whitespace()
 
-            for track in self.tracks:
-                track.metadata_images_changed.connect(self.update_metadata_images)
-                for file in list(track.linked_files):
-                    file.move(self.unmatched_files)
+            unmatched_files = [file for track in self.tracks for file in track.files]
             self.metadata = self._new_metadata
             self.orig_metadata.copy(self.metadata)
+            self.orig_metadata.images.clear()
             self.tracks = self._new_tracks
             del self._new_metadata
             del self._new_tracks
             self.loaded = True
-            self.status = None
-            self.match_files(self.unmatched_files.files)
+            self.status = AlbumStatus.LOADED
+            self.match_files(unmatched_files + self.unmatched_files.files)
             self.enable_update_metadata_images(True)
+            self.update_metadata_images()
             self.update()
             self.tagger.window.set_statusbar_message(
                 N_('Album %(id)s loaded: %(artist)s - %(album)s'),
@@ -413,7 +427,7 @@ class Album(DataObject, Item):
             if self.item.isSelected():
                 self.tagger.window.refresh_metadatabox()
 
-    def _finalize_loading_track(self, track_node, metadata, artists, va, absolutetracknumber, discpregap):
+    def _finalize_loading_track(self, track_node, metadata, artists, extra_metadata=None):
         # As noted in `_parse_release` above, the release artist nodes
         # may contain supplementary data that isn't present in track
         # artist nodes. Similarly, the track artists may contain
@@ -430,17 +444,12 @@ class Album(DataObject, Item):
         tm = track.metadata
         tm.copy(metadata)
         track_to_metadata(track_node, track)
-        tm["~absolutetracknumber"] = absolutetracknumber
         track._customize_metadata()
 
         self._new_metadata.length += tm.length
         artists.add(tm["artist"])
-        if va:
-            tm["compilation"] = "1"
-        else:
-            del tm["compilation"]
-        if discpregap:
-            tm["~discpregap"] = "1"
+        if extra_metadata:
+            tm.update(extra_metadata)
 
         # Run track metadata plugins
         try:
@@ -448,7 +457,6 @@ class Album(DataObject, Item):
         except BaseException:
             self.error_append(traceback.format_exc())
 
-        track.orig_metadata.copy(tm)
         return track
 
     def load(self, priority=False, refresh=False):
@@ -460,20 +468,21 @@ class Album(DataObject, Item):
             {'id': self.id}
         )
         self.loaded = False
-        self.status = _("[loading album information]")
+        self.status = AlbumStatus.LOADING
         if self.release_group:
             self.release_group.loaded = False
             self.release_group.genres.clear()
         self.metadata.clear()
         self.genres.clear()
-        self.update()
+        self.update(update_selection=False)
         self._new_metadata = Metadata()
         self._new_tracks = []
         self._requests = 1
-        self.errors = []
+        self.clear_errors()
+        config = get_config()
         require_authentication = False
         inc = ['release-groups', 'media', 'discids', 'recordings', 'artist-credits',
-               'artists', 'aliases', 'labels', 'isrcs', 'collections']
+               'artists', 'aliases', 'labels', 'isrcs', 'collections', 'annotation']
         if self.tagger.webservice.oauth_manager.is_authorized():
             require_authentication = True
             inc += ['user-collections']
@@ -500,28 +509,37 @@ class Album(DataObject, Item):
             self.tagger.webservice.remove_task(self.load_task)
             self.load_task = None
 
-    def update(self, update_tracks=True):
+    def update(self, update_tracks=True, update_selection=True):
         if self.item:
-            self.item.update(update_tracks)
+            self.item.update(update_tracks, update_selection=update_selection)
 
-    def _add_file(self, track, file):
-        self._files += 1
-        self.update(update_tracks=False)
-        add_metadata_images(self, [file])
-        file.metadata_images_changed.connect(self.update_metadata_images)
+    def add_file(self, track, file, new_album=True):
+        self._files_count += 1
+        if new_album:
+            self.update(update_tracks=False)
+            add_metadata_images(self, [file])
 
-    def _remove_file(self, track, file):
-        self._files -= 1
-        self.update(update_tracks=False)
-        file.metadata_images_changed.disconnect(self.update_metadata_images)
-        remove_metadata_images(self, [file])
+    def remove_file(self, track, file, new_album=True):
+        self._files_count -= 1
+        if new_album:
+            self.update(update_tracks=False)
+            remove_metadata_images(self, [file])
 
-    def _match_files(self, files, recordingid=None, threshold=0):
+    @staticmethod
+    def _match_files(files, tracks, unmatched_files, threshold=0, use_events_iter=False):
         """Match files to tracks on this album, based on metadata similarity or recordingid."""
+        if use_events_iter:
+            #  TODO: get rid of this completely at some point
+            events_iter = process_events_iter
+        else:
+            def _events_iter(seq):
+                return seq
+            events_iter = _events_iter
+
         tracks_cache = defaultdict(lambda: None)
 
         def build_tracks_cache():
-            for track in self.tracks:
+            for track in tracks:
                 tm_recordingid = track.orig_metadata['musicbrainz_recordingid']
                 tm_tracknumber = track.orig_metadata['tracknumber']
                 tm_discnumber = track.orig_metadata['discnumber']
@@ -532,12 +550,13 @@ class Album(DataObject, Item):
                     tracks_cache[tup] = track
 
         SimMatchAlbum = namedtuple('SimMatchAlbum', 'similarity track')
+        no_match = SimMatchAlbum(similarity=-1, track=unmatched_files)
 
         for file in list(files):
             if file.state == File.REMOVED:
                 continue
             # if we have a recordingid to match against, use that in priority
-            recid = recordingid or file.metadata['musicbrainz_recordingid']
+            recid = file.match_recordingid or file.metadata['musicbrainz_recordingid']
             if recid and mbid_validate(recid):
                 if not tracks_cache:
                     build_tracks_cache()
@@ -552,30 +571,29 @@ class Album(DataObject, Item):
 
             # try to match by similarity
             def candidates():
-                for track in process_events_iter(self.tracks):
-                    yield SimMatchAlbum(
-                        similarity=track.metadata.compare(file.orig_metadata),
-                        track=track
-                    )
+                for track in events_iter(tracks):
+                    similarity = track.metadata.compare(file.orig_metadata)
+                    if similarity >= threshold:
+                        yield SimMatchAlbum(similarity=similarity, track=track)
 
-            no_match = SimMatchAlbum(similarity=-1, track=self.unmatched_files)
             best_match = find_best_match(candidates, no_match)
 
-            if best_match.similarity < threshold:
-                yield (file, no_match.track)
-            else:
-                yield (file, best_match.result.track)
+            yield (file, best_match.result.track)
 
-    def match_files(self, files, recordingid=None):
+    def match_files(self, files):
         """Match and move files to tracks on this album, based on metadata similarity or recordingid."""
-        moves = self._match_files(files,
-                                  recordingid=recordingid,
-                                  threshold=config.setting['track_matching_threshold'])
-        for file, target in moves:
-            file.move(target)
+        if self.loaded:
+            config = get_config()
+            threshold = config.setting['track_matching_threshold']
+            moves = self._match_files(files, self.tracks, self.unmatched_files, threshold=threshold)
+            for file, target in moves:
+                file.move(target)
+        else:
+            for file in list(files):
+                file.move(self.unmatched_files)
 
     def can_save(self):
-        return self._files > 0
+        return self._files_count > 0
 
     def can_remove(self):
         return True
@@ -593,23 +611,22 @@ class Album(DataObject, Item):
         return True
 
     def can_view_info(self):
-        return self.loaded or self.errors
+        return self.loaded or bool(self.errors)
+
+    def can_extract(self):
+        return any(track.can_extract() for track in self.tracks)
 
     def is_album_like(self):
         return True
 
     def get_num_matched_tracks(self):
-        num = 0
-        for track in self.tracks:
-            if track.is_linked():
-                num += 1
-        return num
+        return sum(1 for track in self.tracks if track.is_linked())
 
     def get_num_unmatched_files(self):
         return len(self.unmatched_files.files)
 
     def get_num_total_files(self):
-        return self._files + len(self.unmatched_files.files)
+        return self._files_count + len(self.unmatched_files.files)
 
     def is_complete(self):
         if not self.tracks:
@@ -617,62 +634,39 @@ class Album(DataObject, Item):
         for track in self.tracks:
             if not track.is_complete():
                 return False
-        if self.get_num_unmatched_files():
-            return False
-        else:
-            return True
+        return not self.get_num_unmatched_files()
 
     def is_modified(self):
-        if self.tracks:
-            for track in self.tracks:
-                for file in track.linked_files:
-                    if not file.is_saved():
-                        return True
-        return False
+        return any(self._iter_unsaved_files())
 
     def get_num_unsaved_files(self):
-        count = 0
-        for track in self.tracks:
-            for file in track.linked_files:
-                if not file.is_saved():
-                    count += 1
-        return count
+        return sum(1 for file in self._iter_unsaved_files())
+
+    def _iter_unsaved_files(self):
+        yield from (file for file in self.iterfiles(save=True) if not file.is_saved())
 
     def column(self, column):
         if column == 'title':
-            if self.status is not None:
-                title = self.status
+            if self.status == AlbumStatus.LOADING:
+                title = _("[loading album information]")
+            elif self.status == AlbumStatus.ERROR:
+                title = _("[could not load album %s]") % self.id
             else:
                 title = self.metadata['album']
-            if self.tracks:
-                linked_tracks = 0
-                for track in self.tracks:
-                    if track.is_linked():
-                        linked_tracks += 1
 
-                text = '%s\u200E (%d/%d' % (title, linked_tracks, len(self.tracks))
+            if self.tracks:
+                elems = ['%d/%d' % (self.get_num_matched_tracks(), len(self.tracks))]
                 unmatched = self.get_num_unmatched_files()
                 if unmatched:
-                    text += '; %d?' % (unmatched,)
+                    elems.append('%d?' % (unmatched,))
                 unsaved = self.get_num_unsaved_files()
                 if unsaved:
-                    text += '; %d*' % (unsaved,)
-                # CoverArt.set_metadata uses the orig_metadata.images if metadata.images is empty
-                # in order to show existing cover art if there's no cover art for a release. So
-                # we do the same here in order to show the number of images consistently.
-                if self.metadata.images:
-                    metadata = self.metadata
-                else:
-                    metadata = self.orig_metadata
+                    elems.append('%d*' % (unsaved,))
+                ca_detailed = self.cover_art_description_detailed()
+                if ca_detailed:
+                    elems.append(ca_detailed)
 
-                number_of_images = len(metadata.images)
-                if getattr(metadata, 'has_common_images', True):
-                    text += ngettext("; %i image", "; %i images",
-                                     number_of_images) % number_of_images
-                else:
-                    text += ngettext("; %i image not in all tracks", "; %i different images among tracks",
-                                     number_of_images) % number_of_images
-                return text + ')'
+                return '%s\u200E (%s)' % (title, '; '.join(elems))
             else:
                 return title
         elif column == '~length':
@@ -687,6 +681,8 @@ class Album(DataObject, Item):
             return self.metadata['~totalalbumtracks']
         elif column == 'discnumber':
             return self.metadata['totaldiscs']
+        elif column == 'covercount':
+            return self.cover_art_description()
         else:
             return self.metadata[column]
 
@@ -711,9 +707,9 @@ class Album(DataObject, Item):
         if not self.update_metadata_images_enabled:
             return
 
-        update_metadata_images(self)
-
-        self.update(False)
+        if update_metadata_images(self):
+            self.update(False)
+            self.metadata_images_changed.emit()
 
     def keep_original_images(self):
         self.enable_update_metadata_images(False)
@@ -732,17 +728,18 @@ class NatAlbum(Album):
         self.loaded = True
         self.update()
 
-    def update(self, update_tracks=True):
+    def update(self, update_tracks=True, update_selection=True):
+        config = get_config()
         self.enable_update_metadata_images(False)
         old_album_title = self.metadata["album"]
         self.metadata["album"] = config.setting["nat_name"]
         for track in self.tracks:
             if old_album_title == track.metadata["album"]:
                 track.metadata["album"] = self.metadata["album"]
-            for file in track.linked_files:
+            for file in track.files:
                 track.update_file_metadata(file)
         self.enable_update_metadata_images(True)
-        super().update(update_tracks)
+        super().update(update_tracks, update_selection)
 
     def _finalize_loading(self, error):
         self.update()
